#include "../../version.h"
#if (defined(RL_TOOLS_DISABLE_INCLUDE_GUARDS) || !defined(RL_TOOLS_OPERATIONS_DUMMY_GROUP_3_H)) && (RL_TOOLS_USE_THIS_VERSION == 1)
#pragma once
#define RL_TOOLS_OPERATIONS_DUMMY_GROUP_3_H
#if defined(RL_TOOLS_DISABLE_INCLUDE_GUARDS) || !defined(RL_TOOLS_OPERATIONS_DUMMY_GROUP_3)
    #define RL_TOOLS_OPERATIONS_DUMMY_GROUP_3
    #include "../../containers/operations_dummy.h"
#else
    #error "Group 3 already imported"
#endif
#endif
